
---
{% include 'markdown_content_block_header.j2' %}
{% for block in content_block["collapse"] %}
  {% set collapse_block_loop = loop %}
  {% set content_block_id = content_block_id ~ "-collapse-item-" ~ collapse_block_loop.index %}
  {{ block | render_content_block(content_block_id=content_block_id) }}
{% endfor %}
---
